from argparse import ArgumentParser
from collections import defaultdict
from io import StringIO
from pathlib import Path
from numpy import nan
import pandas as pd
from shutil import rmtree
import typing as T
import yaml

from parser import parse_mother_request


df = None

def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        help="Path for workflow. Serves as input and output.",
        type=str,
    )
    parser.add_argument(
        "--flat",
        action="store_true",
        help=(
            "Write application yaml as a single file, "
            "as opposed to in separate files (opa.yml, gsv.yml, meta.yml) "
            "in a same directory."
        ),
    )
    parser.add_argument(
        "--no-abstract",
        action="store_true",
        help=(
            "Do not abstract variables, i.e. have a directory per application variable, "
            "as opposed to having a single `gsv.yml` and `opa.yml` file with abstracted (common) "
            "variables as header"
        ),
    )
    parser.add_argument(
        "--clear",
        action="store_true",
        help="Remove content of created directories if they already existed."
    )
    return parser


def _row_to_dict(row: pd.Series, prefix: str = ""):
    # Pandas row to dict
    dct = row.to_dict()
    # Remove unused fields
    dct = {key: value for key, value in dct.items() if value is not nan}
    # Remove prefix from keys
    if prefix:
        dct = {key.removeprefix(prefix): value for key, value in dct.items()}
    return dct


def create_yaml_dicts(path: Path):
    global df  # pylint: disable=W0603

    mother_request = path / "conf" / "mother_request.yml"
    df = parse_mother_request(mother_request)
    opa_columns = [col for col in df.columns if col.startswith("opa_")]
    gsv_columns = [col for col in df.columns if col.startswith("gsv_")]
    dcts = defaultdict(dict)
    for app, app_df in df.groupby("Application"):
        for i, (_, request) in enumerate(app_df.iterrows(), start=1):
            dcts[app][i] = {
                    "GSVREQUEST": _row_to_dict(request[gsv_columns], "gsv_"),
                    "OPAREQUEST": _row_to_dict(request[opa_columns], "opa_"),
            }
    return dcts


def write_yamls_flat(path: Path):
    yamls = create_yaml_dicts(path)
    yamls_path = path / "conf" / "applications"
    for app, yaml_content in yamls.items():
        yaml_filename = f"{app.lower()}.yml"
        yaml_filepath = yamls_path / yaml_filename
        yaml_io = StringIO()
        yaml.safe_dump(yaml_content, yaml_io, default_flow_style=False)
        with open(yaml_filepath, "w", encoding="utf-8") as wf:
            wf.write(
                "---\n\n"
            )
            wf.write(
                yaml_io.getvalue().replace("\'", "\"")
            )


def write_yamls_tree(path: Path, clear: bool = False):
    yamls = create_yaml_dicts(path)
    conf_apps_path = path / "conf" / "applications"
    request_filename = [
        ("GSVREQUEST", "gsv.yml"),
        ("OPAREQUEST", "opa.yml"),
    ]
    for app, app_requests in yamls.items():
        app_path = conf_apps_path / app.lower()
        if request_path.exists() and clear:
            rmtree(app_path)
        app_path.mkdir(exist_ok=True, parents=False)
        print(f"Made dir: {app_path}")
        for i, request in app_requests.items():
            request_path = app_path / f"{i:02d}"
            request_path.mkdir(exist_ok=True, parents=False)
            for req, filename in request_filename:
                with open(request_path / filename, "w", encoding="utf-8") as wf:
                    print(request[req])
                    yaml.dump(request[req], wf)
        with open(app_path / "meta.yml", "w", encoding="utf-8") as wf:
            yaml.dump({
                "APPNAME": app,
            })


def _len_unique(values: T.Sequence[T.Any]):
    if any(isinstance(element, list) for element in values):
        values = [
            (element if not isinstance(element, list) else tuple(element))
            for element in values
        ]
    return len(set(values))


def _sort_by_count(lst):
    def sort(x):
        return lst.count(x)
    return sorted(lst, key=sort)


def write_yamls_abstract(
    path: Path,
    clear: bool = False,
    toplevel: bool = True,
):
    global df  # pylint: disable=W0603

    def dump(df, path, toplevel: T.Optional[T.Sequence[str]] = None):  # pylint: disable=W0621
        df = df.dropna(axis=1, how="all")
        unique_variables = [column for column in df.columns if _len_unique(df[column]) == 1]
        non_unique_variables = [column for column in df.columns if column not in unique_variables]
        if unique_variables:
            content = df[unique_variables].to_dict(orient="records")[0]
        else:
            content = {}
        if non_unique_variables:
            non_unique_content = df[non_unique_variables].to_dict(orient="records")
            # Drop originally non-existing keys (lists don't like pd.isna, so evaluate to True first)
            non_unique_content = [
                {key: value for key, value in entry.items() if (isinstance(value, list) or not pd.isna(value))}
                for entry
                in non_unique_content
            ]
            content.update(
                dict(enumerate(non_unique_content, start=1))
            )
        if toplevel:
            if isinstance(toplevel, str):
                toplevel = [toplevel]
            if isinstance(toplevel, (list, tuple)):
                for lev in toplevel[::-1]:
                    content = {lev: content}
            else:
                raise TypeError(f"Bad type for typlevel: {type(toplevel).__name__}.")
        with open(path, "w", encoding="utf-8") as wf:
            yaml.dump(content, wf, sort_keys=False)

    mother_request = path / "conf" / "mother_request.yml"
    conf_apps_path = path / "conf" / "applications"

    df = parse_mother_request(mother_request)
    opa_columns = [col for col in df.columns if col.startswith("opa_")]
    gsv_columns = [col for col in df.columns if col.startswith("gsv_")]

    for app, app_df in df.groupby("Application"):

        app_path = conf_apps_path / app.lower()
        if app_path.exists() and clear:
            rmtree(app_path)
        app_path.mkdir(exist_ok=True, parents=False)

        opa_requests = app_df[opa_columns]
        opa_requests.columns = [col.removeprefix("opa_") for col in opa_columns]
        gsv_requests = app_df[gsv_columns]
        gsv_requests.columns = [col.removeprefix("gsv_") for col in gsv_columns]

        dump(
            opa_requests,
            app_path / "opa.yml",
            toplevel=(app, "OPAREQUEST") if toplevel else None
        )
        dump(
            gsv_requests,
            app_path / "gsv.yml",
            toplevel=(app, "GSVREQUEST") if toplevel else None
        )

        with open(app_path / "meta.yml", "w", encoding="utf-8") as wf:
            yaml.dump({
                "APPNAME": app,
            }, wf, sort_keys=False)


if __name__ == "__main__":
    args = _get_parser().parse_args()

    if args.flat:
        write_yamls_flat(Path(args.path))
    else:
        if args.no_abstract:
            write_yamls_tree(Path(args.path))
        else:
            write_yamls_abstract(Path(args.path), args.clear)
